library(tidyverse)

options(mc.cores = parallel::detectCores())

source("lib.R")

pruned <- "estimated_rates/pruned" %>% file.path("*.raxml.bestModel") %>% Sys.glob()
reanalyzed <- "estimated_rates/raxml" %>% file.path("*.raxml.bestModel") %>% Sys.glob()

read_raxml_ng_model_parameters <- function(x) {
    # Rate matrix: AC AG AT CG CT GT
    # Frequencies: A C G T
    l <- readLines(x)
    l <- str_extract(l, "[^,]+")
    pieces <- str_split_fixed(l, fixed("+"), 3)
    gtr <- as.double(str_extract_all(pieces[1], "[0-9.]+")[[1]])
    freqs <- as.double(str_extract_all(pieces[2], "[0-9.]+")[[1]])
    alpha <- as.double(str_extract_all(pieces[3], "[0-9.]+")[[1]][2])

    aa <- c(gtr, freqs, alpha)
    names(aa) <- str_split("AC AG AT CG CT GT A C G T alpha", " ")[[1]]
    df <- as_data_frame(t(aa))
    df
}

fx <- function(x) {
    read_raxml_ng_model_parameters(x) %>% mutate(family = str_extract(basename(x), "[^.]+"))
}

pruned_df <- parallel::mclapply(pruned, fx) %>% bind_rows() %>%  gather("variable", "pruned", -family)
reanalyzed_df <- parallel::mclapply(reanalyzed, fx) %>% bind_rows() %>%  gather("variable", "reanalyzed", -family)

reference <- read_raxml_ng_model_parameters("estimated_rates/actinopt.phy.raxml.bestModel") %>% gather("variable", "reference")

all_df <- left_join(pruned_df, reanalyzed_df) %>% left_join(reference) %>% mutate(diff_pruned = pruned - reference, diff_reanalyzed = reanalyzed - reference)

fmted <- select(all_df, family, variable, diff_pruned, diff_reanalyzed) %>% gather("key", "value", -family, -variable) %>% filter(key == "diff_reanalyzed", variable != "GT", value < 50)

ggplot(fmted, aes(x = value)) + geom_histogram() + facet_wrap(~variable, scales = "free")

au <- read_csv("au_results.csv")
dr <- read_csv("fulltrees_dr_summary.csv")

wide <- fmted %>% select(family, variable, value) %>% spread(variable, value)

test <- left_join(au, wide)

ggplot(left_join(au, fmted), aes(x = shape, y = value, color = variable)) + geom_point() + facet_wrap(~variable, scales = "free") + theme(legend.position = "none")

res <- left_join(pruned_df, au) %>% filter(alpha < 10)
res2 <- left_join(reanalyzed_df, au) %>% filter(alpha < 10)

aa <- res %>% group_by(family) %>% summarise(tl_pruned = median(tl))
aa2 <- res2 %>% group_by(family) %>% summarise(tl_reanalyzed = median(tl))

asdf <- left_join(aa, aa2) %>% mutate(diff = (tl_pruned - tl_reanalyzed) / tl_reanalyzed) %>% left_join(au)

gene <- lm(tl ~ (a + ac + ag + at + c + ct + g + alpha + patterns) * gene, data = res)
no_gene <- lm(tl ~ (a + ac + ag + at + c + ct + g + alpha + patterns), data = res)
anova(gene, no_gene)
# no gene is supported

res %>% distinct(family, a, ac, ag, at, c, ct, g, alpha, patterns) %>% predict(no_gene, .)
