#!/usr/bin/env Rscript

library(magrittr)
library(ape)
library(stringr)
library(dplyr)
library(ggplot2)

fn <- "output/pruned_trees/Acanthuridae.tre"
fn <- commandArgs(T)

RAXML <- "raxmlHPC-AVX2"

SEQPATH <- "raxml/"

OUTPATH <- "output/au_test/"
dir.create(OUTPATH, recursive = T, showWarnings = F)

old <- read.tree(fn)
new <- str_replace(fn, "pruned_trees", "raxml_trees") %>% read.tree
if (class(new) == "multiPhylo")
    new <- new[[length(new)]]
res <- list(old, new)
class(res) <- "multiPhylo"


family <- basename(fn) %>% str_replace(".tre$", "")

wd <- tempdir()
wd2 <- file.path(wd, family)
dir.create(wd2)
write.tree(res, file.path(wd2, "all.trees"))
file.copy(file.path(SEQPATH, family, "dat.phy"), wd2)
file.copy(file.path(SEQPATH, family, "dat.partitions"), wd2)
system(paste(RAXML, "-f c -w", wd2, "-s", file.path(wd2, "dat.phy"), "-q", file.path(wd2, "dat.partitions"), "-m GTRGAMMA -T 2 -n red"))
if (file.exists(file.path(wd2, "dat.phy.reduced"))) {
    file.rename(file.path(wd2, "dat.phy.reduced"), file.path(wd2, "dat.phy"))
    file.rename(file.path(wd2, "dat.partitions.reduced"), file.path(wd2, "dat.partitions"))
}
system(paste(RAXML, "-f G -w", wd2, "-s", file.path(wd2, "dat.phy"), "-q", file.path(wd2, "dat.partitions"), "-z", file.path(wd2, "all.trees"), "-m GTRGAMMA -n consel"))
system(paste("makermt -b 10 --puzzle", file.path(wd2, "RAxML_perSiteLLs.consel")))
system(paste("consel", file.path(wd2, "RAxML_perSiteLLs")))
system(paste("catpv -e", file.path(wd2, "RAxML_perSiteLLs"), ">", file.path(OUTPATH, paste0(family, ".txt"))))


q(status = 0)



# AU test


# RF distances
distances <- sapply(trees, function(x) RF.dist(x)[[1]])
sizes <- sapply(trees, function(x) length(x[[1]]$tip.label))

df <- data_frame(family = names(distances), dist = distances[family], ntip = sizes[family])

ggplot(df, aes(ntip, dist, label = family)) + geom_jitter() + geom_text(data = subset(df, ntip > 400 | dist > 50), hjust=1.1, size=3)  + coord_trans(x = "log10") + scale_x_continuous(breaks = c(10, 100, 1000)) + theme_minimal() + xlab("Number of tips") + ylab("RF-distance")


pml_fits <- lapply(trees, function(tt) {

})

RF.dist(trees[[1]])
